# Copyright 2022 ECCO Sneaks & Data
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A union find data structure for collecting results of the anti-clustering algorithm.

Based on:
Sedgewick, R. & Wayne, K. (2011), Algorithms, 4th Edition. , Addison-Wesley .
"""

from typing import TypeVar, Generic

T = TypeVar("T")  # pylint: disable=C0103


class UnionFind(Generic[T]):
    """
    A union find data structure for collecting results of the anti-clustering algorithm.
    This implementation uses the weighted quick union with path compression.
    """

    # A mapping from an element to its parent. If a parent maps to itself, it is the root of the component.
    _parent = {}
    _size = {}
    components_count = 0

    def __init__(self, initial_components_count: int):
        """
        Initialize UnionFind with components.
        :param initial_components_count: The initial number of components.
        """
        self.components_count = initial_components_count
        self._parent = {i: i for i in range(initial_components_count)}
        self._size = {i: 1 for i in range(initial_components_count)}

    def _find(self, element: T) -> T:
        """
        Find the root of component of element.
        :param element: Element to find root of.
        :return: The root of the component.
        """
        # Compresses path while iterating up the tree.
        while element != self._parent[element]:
            parent = self._parent[element]
            self._parent[element] = self._parent[parent]
            element = parent

        return element

    def find(self, element: T) -> T:
        """
        Find the root of component of element.
        :param element: Element to find root of.
        :return: The root of the component.
        """
        return self._find(element)

    def union(self, element_1: T, element_2: T) -> None:
        """
        Unify components of two elements.
        :param element_1: Element to unify.
        :param element_2: Other element to unify.
        :return:
        """
        if element_1 == element_2:
            return

        root_1 = self._find(element_1)
        root_2 = self._find(element_2)

        if root_1 == root_2:
            return

        # Weighted union - the smaller component becomes the child of the root of the larger component.
        if self._size[root_1] < self._size[root_2]:
            self._parent[root_1] = root_2
            self._size[root_2] += self._size[root_1]
        else:
            self._parent[root_2] = root_1
            self._size[root_1] += self._size[root_2]
        self.components_count -= 1

    def connected(self, element_1: T, element_2: T) -> bool:
        """
        Check if element 1 and 2 are in the same component.
        :param element_1: Element to check.
        :param element_2: Other element to check.
        :return: Whether 1 and 2 are in the same component.
        """
        return self._find(element_1) == self._find(element_2)



# Copyright 2022 ECCO Sneaks & Data
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract class containing utilities for cluster swap-based heuristics."""

from abc import ABC
import random
import numpy as np
import numpy.typing as npt
from anti_clustering._base import AntiClustering
from anti_clustering._union_find import UnionFind


class ClusterSwapHeuristic(AntiClustering, ABC):
    """Abstract class containing utilities for cluster swap-based heuristics."""

    def __init__(self, verbose: bool = False, random_seed: int = None):
        super().__init__(verbose=verbose)
        self.rnd = random.Random(random_seed)

    def _get_exchanges(self, cluster_assignment: npt.NDArray[bool], i: int) -> npt.NDArray[int]:
        """
        Given a cluster assignment matrix and element index, will return possible indexes to swap anti-clusters with.
        :param cluster_assignment: Cluster assignment matrix.
        :param i: Element index.
        :return: Possible exchanges.
        """
        return np.nonzero(np.invert(cluster_assignment[i]))[0]

    def _swap(self, cluster_assignment: npt.NDArray[bool], i: int, j: int) -> npt.NDArray[bool]:
        """
        Swap anti-clusters of elements i and j.
        :param cluster_assignment: Current cluster assignment.
        :param i: Element.
        :param j: Other element.
        :return: Cluster assignment with i and j swapped.
        """
        cluster_assignment = cluster_assignment.copy()
        tmp1 = cluster_assignment[i,].copy()
        tmp2 = cluster_assignment[:, i].copy()
        cluster_assignment[i,] = cluster_assignment[j,]
        cluster_assignment[:, i] = cluster_assignment[:, j]
        cluster_assignment[j,] = tmp1
        cluster_assignment[:, j] = tmp2
        cluster_assignment[i, j] = False
        cluster_assignment[j, i] = False
        return cluster_assignment

    def _get_random_clusters(self, num_groups: int, num_elements: int) -> npt.NDArray[bool]:
        """
        Get a random initialization of anti-clusters.
        :param num_groups: Number of anti-clusters to generate.
        :param num_elements: Number of elements in algorithm run.
        :return: The randomly initialized anti-clusters as an assignment matrix
        """
        if self.verbose:
            print("Initializing clusters")

        # Using UnionFind to generate random anti-clusters. The first num_groups elements are guaranteed
        # to be roots of each their own component. All other elements are assigned a random root.
        initial_clusters = [i % num_groups for i in range(num_elements - num_groups)]
        self.rnd.shuffle(initial_clusters)
        initial_clusters = list(range(num_groups)) + initial_clusters
        uf_init = UnionFind(len(initial_clusters))  # pylint: disable = R1721
        for i, cluster in enumerate(initial_clusters):
            uf_init.union(i, cluster)

        cluster_assignment = np.array(
            [[uf_init.connected(i, j) for i in range(num_elements)] for j in range(num_elements)]
        )

        return cluster_assignment

    def _calculate_objective(self, cluster_assignment: npt.NDArray[bool], distance_matrix: npt.NDArray[float]) -> float:
        """
        Calculate objective value
        :param cluster_assignment: Cluster assignment
        :param distance_matrix: Distance matrix
        :return: Objective value
        """
        return np.multiply(cluster_assignment, distance_matrix).sum()



# Copyright 2022 ECCO Sneaks & Data
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A tabu search with restarts approach to solving the anti-clustering problem.
"""

import numpy.typing as npt
from anti_clustering._cluster_swap_heuristic import ClusterSwapHeuristic


class TabuSearchHeuristicAntiClustering(ClusterSwapHeuristic):
    """
    A tabu search with restarts approach to solving the anti-clustering problem.
    In this version, specific transformations are put in the tabu list not solutions.
    """

    def __init__(
        self,
        verbose: bool = False,
        random_seed: int = None,
        tabu_tenure: int = 10,
        iterations: int = 2000,
        restarts: int = 9,
    ):
        # pylint: disable = R0913
        super().__init__(verbose=verbose, random_seed=random_seed)
        self.tabu_tenure = tabu_tenure
        self.iterations = iterations
        self.restarts = restarts

    def _solve(self, distance_matrix: npt.NDArray[float], num_groups: int) -> npt.NDArray[bool]:
        # Start with random cluster assignment
        cluster_assignment = self._get_random_clusters(num_groups=num_groups, num_elements=len(distance_matrix))

        if self.verbose:
            print("Solving")

        candidate_solutions = []

        for restart in range(self.restarts):
            tabu_swaps = []
            # Initial objective value
            objective = self._calculate_objective(cluster_assignment, distance_matrix)
            for iteration in range(self.iterations):
                if self.verbose and iteration % 5 == 0:
                    print(f"Iteration {iteration + 1} of {self.iterations}")

                # Select random element
                i = self.rnd.randint(0, len(distance_matrix) - 1)

                # Get possible swaps
                possible_exchanges = [
                    j
                    for j in self._get_exchanges(cluster_assignment, i)
                    if (i, j) not in tabu_swaps and (j, i) not in tabu_swaps
                ]

                if len(possible_exchanges) == 0:
                    continue

                # Generate possible assignments
                j = possible_exchanges[self.rnd.randint(0, len(possible_exchanges) - 1)]

                # Select random possible swap.
                new_cluster_assignment = self._swap(cluster_assignment, i, j)
                new_objective = self._calculate_objective(new_cluster_assignment, distance_matrix)

                # Select solution as current if it improves the objective value
                if new_objective > objective:
                    cluster_assignment = new_cluster_assignment
                    objective = new_objective
                    tabu_swaps.append((i, j))
                    # Delete oldest tabu swap if tabu list is full
                    if len(tabu_swaps) > self.tabu_tenure:
                        tabu_swaps.pop(0)

            candidate_solutions.append((objective, cluster_assignment))

            if self.verbose:
                print(f"Restart {restart + 1} of {self.restarts}")

            # Cold restart, select random cluster assignment
            cluster_assignment = self._get_random_clusters(num_groups=num_groups, num_elements=len(distance_matrix))

        # Select best solution, maximizing objective
        _, best_cluster_assignment = max(candidate_solutions, key=lambda x: x[0])

        return best_cluster_assignment



# Copyright 2022 ECCO Sneaks & Data
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The exchange heuristic to solving the anti-clustering problem.

Based on:
Papenberg, M., & Klau, G. W. (2021). Using anticlustering to partition data sets into equivalent parts.
Psychological Methods, 26(2), 161–174. https://doi.org/10.1037/met0000301
"""

import numpy.typing as npt
from anti_clustering._cluster_swap_heuristic import ClusterSwapHeuristic


class ExchangeHeuristicAntiClustering(ClusterSwapHeuristic):
    """
    The exchange heuristic to solving the anti-clustering problem.
    """

    def __init__(self, verbose: bool = False, random_seed: int = None, restarts: int = 9):
        super().__init__(verbose=verbose, random_seed=random_seed)
        self.restarts = restarts

    def _solve(self, distance_matrix: npt.NDArray[float], num_groups: int) -> npt.NDArray[bool]:
        # Starts with random cluster assignment
        cluster_assignment = self._get_random_clusters(num_groups=num_groups, num_elements=len(distance_matrix))

        if self.verbose:
            print("Solving")

        candidate_solutions = []

        for restart in range(self.restarts):
            # Initial objective value
            current_objective = self._calculate_objective(cluster_assignment, distance_matrix)
            for i in range(len(distance_matrix)):
                if self.verbose and i % 5 == 0:
                    print(f"Iteration {i + 1} of {len(distance_matrix)}")

                # Get list of possible swaps
                exchange_indices = self._get_exchanges(cluster_assignment, i)

                if len(exchange_indices) == 0:
                    continue

                # Calculate objective value for all possible swaps.
                # List contains tuples of obj. val. and swapped element index.
                exchanges = [
                    (self._calculate_objective(self._swap(cluster_assignment, i, j), distance_matrix), j)
                    for j in exchange_indices
                ]

                # Find best swap
                best_exchange = max(exchanges)

                # If best swap is better than current objective value then complete swap
                if best_exchange[0] > current_objective:
                    cluster_assignment = self._swap(cluster_assignment, i, best_exchange[1])
                    current_objective = best_exchange[0]

            candidate_solutions.append((current_objective, cluster_assignment))

            if self.verbose:
                print(f"Restart {restart + 1} of {self.restarts}")

            # Cold restart, select random cluster assignment
            cluster_assignment = self._get_random_clusters(num_groups=num_groups, num_elements=len(distance_matrix))

        # Select best solution, maximizing objective
        _, best_cluster_assignment = max(candidate_solutions, key=lambda x: x[0])

        return best_cluster_assignment
